import os

_include_iteration_max_depth = 16

"""
Goal is to be able to run an 'eda' command, such as:
     > eda sim --target-packager <--stop-before-compile> \
               --tool verilator [--waves] [--dump-vcd] [other args] \
               <files or target>

Or a 'multi sim' command:
     > eda multi sim --target-packager <--stop-before-compile> \
               --tool verilator [--waves] [--dump-vcd] [other args] \
               <files or target wildcard>

And doing so would create:
   ./eda.work/<target>_sim/target_packager/
      ./DEPS.yml              -- new file with single target, all defines, incdirs, sources, top, etc.
      ./<all source files>    -- edited to flatten path information in `include
      ./<all `included files> -- edited to flatten path information in `include
      ./test.json             -- single file that can be as starting point for test runner service
                              -- Or a singular test.jsonl

A multi sim command would also create:
   ./eda.work/multi/tests.jsonl

And, we could re-run a single target in place as, but won't support this for 'multi sim':
  > cd ./eda.work/<target>_sim/target_packager/
  > eda sim test

`include problem:
1. In SystemVerilog, a `include "foo.svh", no tools allow you to add this to your compile/build
   flow via filename. They all do this per directory.
   -- As a result, we do not track includes, other than "incdirs" in our DEPS.yml tables.
   -- relative path or absolute path includes are allowed:
      - `include "../../../../usr/bin/foo.svh"
      - `include "deeper/path/lets/go/foo.svh"
2. Given a compile, we do not know the exact individual files were `include'd.
3. Verific will not tell us which files wer actually `include'd in $root scope (would tell us
   within the module scope).
4. WHY IS THIS A PROBLEM
   - If we do nothing, Frogger will be responible for creating all relative or absolute
     directories. I see this as a security risk and infeasible to implement.
     -- aka, Frogger runs a test in /
5. Solutions:
   a. One option is: we make a decision to not support path information in includes, and
      manage this only in DEPS.yml 'incdirs'.
      -- This may not be ideal for some hypothetical customer in the future.
         (If you want to use our tool and Frogger, refactor your code first!)
   b. Another option is - we support limited parsing in this "eda sim --target-packager" flow
      to copy + edit the file to strip path information in includes, so that Frogger can create
      all files in a single working directory.


*IF* we make the decision to only support `include w/out path information in
the opencos repo, and suggest that this is preferred, then <all * files> could
instead be symbolic links.
   ./eda.work/<target>_sim/target_packager/
      ./DEPS.yml              -- new file with single target, all defines, incdirs, sources, top, etc.
      ./<all source files>    -- as symbolic links
      ./<all `included files> -- as symbolic links
      ./test.json             -- single file that can be as starting point for test runner service


*IF* we want to support relative or absolute path based `include "../../foo.svh", then
we get to unravel that with a copied and edited file.
Looking in eth-puzzles/bin/create_tests_jsonl_eda.py, the first step is running (since we didn't
edit eda.py at all) effectively:

     > eda sim --stop-before-compile --tool verilator <some-target>

and examining the eda.work/some-target_sim/compile_only.sh file.# From there,
the commands are split and we figure out incdirs, defines, source files, top
module name and any other knobs.
We then get the full path information of every source file, and infer the 'top'
file if it was not set otherwise. Finally, we resolve the `include nightmare in
all source files, and check that they exist in the incdirs we know about.

** If a file has no `include, and no edits, it will be a symbolic link
** If a file had any edits, it will not be a symbolic link
** We must check that edited files only have affected diff lines on lines containing `include.


An example packaged DEPS.yml file:

test:
  multi:
    ignore:
      # This is required so that generated DEPS.yml files are not picked
      # up by future 'eda multi sim' commands.
      - commands: sim synth build
  tags:
    only-tools:
      # Note, tags are not fully supported by 'eda' yet, but if the
      # original command was spec'd for a verilator compile list, then
      # that should be the only tool supported.
      - verilator
  incdirs: .
  deps:
    - oclib_assert_pkg.sv
    - oclib_pkg.sv
    - oclib_memory_bist_pkg.sv
    - oclib_uart_pkg.sv
    - ocsim_pkg.sv
    - ocsim_packet_pkg.sv
    - ocsim_tb_control.sv
    - oclib_ram1rw.sv
    - oclib_ram1rw_test.sv
  top: oclib_ram1rw_test
  defines:
    LATENCY: 0


An example of a test.json file (w/out ## comments b/c json is a no comments zone):

  "name":    "oclib_am1rw_test",   ## This might match the DEPS.yml - <target>: top value.
                                   ## Drew would expect this to be the how I can identify my test
                                   ## when I want to look a logs, errors, artifacts

  "eda":     true,                 ## Must be true, Frogger will run this using 'eda'

  "eda_command": "eda sim --waves --tool verilator test",
                                   ## Command Frogger should expect to run
                                   ## If we find this to be gross or bad-security-practice
                                   ## then I'm not sure on the alterative. Frogger could
                                   ## be aware of allow-listed args that 'eda' suports, so
                                   ## Chip team putting its desired command in a json Array of
                                   ## args is not any different than a space separated string?
                                   ## recommand: mylist = shlex.split(eda_command)

  "eda_target": "test",            ## Name of the DEPS.yml <target>.
                                   ## Could instead use "tb_name" but this is Drew's preference.

  "tb": [
    {"name": "DEPS.yml",          "content": <string file contents> },
    {"name": oclib_assert_pkg.sv, "content": <string file contents> },
    ## ...
    {"name": oclib_ram1rw_test.sv, "content": <string file contents> },

    ## Note that this will match the same order in the Array in DEPS.yml <target>:deps value.
    ## This is the compile order.

    ## We will not pass other hidden information that Frogger has to deal with. Any
    ## 'defines' or 'incdirs' or other Verilator compile flags will be in the DEPS.yml.
  ],

  ## AI Team may have other exciting test runner Table key/value items. Drew has no use for:
  ## - canonical_dut, dut, query, prefix, top, tb_name (if I get "eda_target")

"""

def find_sv_included_files_within_file(filename:str, known_incdir_paths:list) -> list:

    found_included_files = set()
    ret = list()

    assert any(filename.endswith(x) for x in ['.v', '.sv', '.vh', '.svh']), \
        f'{filename=} does not have a supported extension, refusing to parse it'
    assert os.path.exists(filename), f'{filename=} does not exist'

    with open(filename) as f:

        for line in f.readlines():
            if '`include' in line:
                # strip comments on line, in case someone has: // `include "lib/foo.svh"
                # we can't handle /* comments */ on a line like this.
                assert '/*' not in line
                parts = line.split("//")
                words = parts[0].split() # only use what's on the left of the comments
                prev_word_is_tick_include = False
                ##print(f'(DEBUG): found `include in {line=} {parts=} {words=}')
                for word in words:
                    word = word.rstrip('\n')
                    if word == '`include':
                        # don't print this word, wait until next word
                        prev_word_is_tick_include = True
                        ##print(f'(DEBUG):    {word=}')
                    elif prev_word_is_tick_include:
                        assert word.startswith('"')
                        assert word.endswith('"')
                        prev_word_is_tick_include = False
                        include_fname = word[1:-1] # trim " at start and end
                        if include_fname not in found_included_files:
                            # this has path information, perhaps relative, perhaps absolute, or
                            # perhaps relative to any of the +incdir+ paths. Figure that out later.
                            found_included_files.add(include_fname)

    for fname in found_included_files:
        # Does this file exist, using our known_incdir_paths?
        for some_dir in known_incdir_paths:
            try_file_path = os.path.join(some_dir, fname)
            if os.path.exists(try_file_path) and try_file_path not in ret:
                ret.append(try_file_path)

        # TODO(drew): Warning if file doesn't exist, this would eventually result in a
        # compiler error.
        # TODO(drew): Warning if included file doesn't end  in .v, .sv, .vh, .svh.

    return ret



def get_list_sv_included_files(all_src_files:list, known_incdir_paths:list, target:str='') -> list:

    sv_included_files_dict = dict() # key, value is if we've traversed it (bool)

    # We now have to comb through all the tb_files (aka all src files including dut.sv)
    # so we can store them in an ordered file list for compiling.
    # We do have to go through all the src files looking for includes to flattend the path
    # information:
    #    `include "path/to/foo.s?vh" with
    # needs to become
    #    `include "foo.s?vh"
    # We also have to locate these files path/to/foo.s?vh" and add them to our source files
    # aka tb_files, b/c they were not otherwise in the list from eda's compile_only.sh

    # order shouldn't matter, these will get added to the testrunner's filelist and
    # be included with +incdir+.

    for fname in all_src_files:
        included_files_list = find_sv_included_files_within_file(
            filename=fname,
            known_incdir_paths=known_incdir_paths
        )

        for f in included_files_list:
            if f not in sv_included_files_dict:
                sv_included_files_dict[f] = False # add entry, mark it not traversed.

    for _ in range(_include_iteration_max_depth):
        # TODO(drew): consider doing this as a while not all(sv_included_files_dict.values())
        # (meaning one or more is false, aka not traversed)

        # do these for a few levels, in case `include'd file includes another file.
        # If we have more than 8 levels of `include hunting, then rethink this.
        for fname,traversed in sv_included_files_dict.items():
            if not traversed:
                included_files_list = find_sv_included_files_within_file(
                    filename=fname,
                    known_incdir_paths=known_incdir_paths
                )
                sv_included_files_dict[fname] = True # mark as traversed.

                for f in included_files_list:
                    if f not in sv_included_files_dict:
                        sv_included_files_dict[f] = False # add entry, mark it not traversed.

    ret = list()
    for fname,traversed in sv_included_files_dict.items():
        assert traversed, f'{fname=} has not been traversed, {sv_included_files_dict=} from {target=}'

        # add all the included files (should be traversed!) to our return list
        ret.append(fname)

    return ret



import os

# Given what we'll assume is a CommandDesign (CommandSim?) object, we'll try to
# perform some invio related commands on it.

def write_py_file(command_design_obj:object, py_filename='run_invio.py', v_filename='',
                  blackbox_list=[], **kwargs) -> dict:
    '''Given an eda.CommandDesign object, creates a .py file that can be run

    from invio-py or standalone. Returns a dict with a few details about the generated .py'''


    # Sanity check that command_design_obj has some expected details.
    assert type(getattr(command_design_obj, 'incdirs', None)) is list
    assert type(getattr(command_design_obj, 'files_v', None)) is list
    assert type(getattr(command_design_obj, 'files_sv', None)) is list
    assert type(getattr(command_design_obj, 'files_vhd', None)) is list

    args = getattr(command_design_obj, 'args', dict())
    work_dir = args.get('work-dir', '')

    assert work_dir

    work_dir = os.path.abspath(work_dir)

    lines = [

        'from invio import init, define_macro, add_include_directory, \\',
        '    add_verilog_file, add_sv_file, elaborate, analyze, write_design',
        'import os, shutil',
        '',
        f'for p in ["invio"]:',
        f'    fullp = os.path.join("{work_dir}", p)',
        f'    if not os.path.exists(fullp):',
        f'        os.mkdir(fullp)',
        '',
    ]

    for name,value in command_design_obj.defines.items():
        if name in ['SIMULATION']:
            continue

        if value is None:
            lines.append(f'define_macro("{name}")')
        else:
            value = str(value)
            value = value.replace("'", "\\" + "'")
            value = value.replace('"', "\\" + '"')
            lines.append(f'define_macro("{name}", value="{value}")')

    # We must define SYNTHESIS for oclib_defines.vh to work correctly.
    lines.append(f'define_macro("SYNTHESIS")')

    for path in command_design_obj.incdirs:
        lines.append(f'add_include_directory("{path}")')

    for path in command_design_obj.files_v:
        lines.append(f'add_verilog_file("{path}")')

    for path in command_design_obj.files_sv:
        lines.append(f'add_sv_file("{path}")')

    for path in command_design_obj.files_vhd:
        lines.append(f'add_vhdl_file("{path}")')

    blackbox_files_list = list()
    for path in command_design_obj.files_v + command_design_obj.files_sv:
        leaf_filename = path.split('/')[-1]
        module_name = ''.join(leaf_filename.split('.')[:-1])
        if module_name in blackbox_list:
            blackbox_files_list.append(path)


    top = args.get('top', None)
    if top is None:
        # Infer the 'top' module as last verilog, SV, or vhdl file.
        tmp_list = command_design_obj.files_v + command_design_obj.files_sv \
            + command_design_obj.files_vhd
        top_fileonly = tmp_list[-1].split('/')[-1]
        top = ''.join(top_fileonly.split('.')[:-1])


    assert top

    if v_filename:
        v_filename = v_filename
    else:
        v_filename = f'{top}.v'


    full_v_filename = os.path.join(work_dir, 'invio', v_filename)
    full_py_filename = os.path.join(work_dir, py_filename)

    ret = {
        'work_dir': work_dir,
        'py_filename': py_filename,
        'v_filename': v_filename,
        'full_v_filename': full_v_filename,
        'full_py_filename': full_py_filename,
        'top': top,
        'blackbox_module_list': blackbox_list,
        'blackbox_files_list': blackbox_files_list,
    }



    lines += [
        'assert analyze()',
        f"assert elaborate('{top}', rtl_elaboration=True, forceBlackbox={blackbox_list})",
        '',
        f"assert write_design('{full_v_filename}')",
        '',
    ]

    # It seems that Invio's write_design(..) will leave a few VERIFIC_MUX and other modules
    # commented out in the .v file. This doesn't work downstream for yosys, so uncomment these
    # with some brute force.

    tmp_v_filename = os.path.join(work_dir, '_tmp.v')

    lines += [
        '',
        '# We also get the joy of uncommenting any VERIFIC_* modules in the generated .v',
        f'with open("{full_v_filename}") as old, open("{tmp_v_filename}", "w") as new:',
        '    in_verific_commented_module = False',
        '    for line in old.readlines():',
        '        # Handle old, modify line:',
        '        if "// module VERIFIC_" in line:',
        '            in_verific_commented_module = True',
        '        if in_verific_commented_module:',
        '            line = line.replace("// ", "")',
        '        if line.startswith("endmodule"):',
        '            in_verific_commented_module = False',
        '        # Write out line to new:',
        '        new.write(line)',
        f'shutil.move("{tmp_v_filename}", ',
        f'            "{full_v_filename}")',
        '',
    ]


    for k,v in ret.items():
        if type(v) is str:
            lines.append(f'## {k} = "{v}"')


    if py_filename:
        outfile = os.path.join(work_dir, py_filename)
        with open(outfile, 'w') as f:
            f.write('\n'.join(lines))


    return ret




def write_verilog(command_design_obj:object, v_filename='', py_filename='run_invio.py',
                  blackbox_list=[], **kwargs) -> dict:
    '''Creates and runs a .py file to run Invio on a CommandDesign oject. Returns a dict with

    information about what was run.'''

    invio_dict = write_py_file( command_design_obj, v_filename=v_filename,
                                py_filename=py_filename, blackbox_list=blackbox_list,
                                **kwargs )

    # exec the python that was created:
    # Do not run this if args['stop-before-compile'] is True:
    if command_design_obj.args.get('stop-before-compile', False):
        pass # skip it.
    else:
        if 'full_py_filename' in invio_dict:
            exec(open(invio_dict['full_py_filename']).read())

    return invio_dict




# SPDX-License-Identifier: MPL-2.0

import sys
import subprocess
import datetime
import os
import time
import atexit
import shutil
import traceback
import argparse

# Conditional imports for loading this as a module instead of package
# where someone may not have pyyaml installed.
try:
    import yaml
except:
    pass

global_exit_allowed = False
progname = "UNKNOWN"
progname_in_message = True
logfile = None
loglast = 0
debug_level = 0

args = {
    'color' : True,
    'quiet' : False,
    'verbose' : False,
    'debug' : False,
    'fancy' : sys.stdout.isatty(),
    'warnings' : 0,
    'errors' : 0,
}

def yaml_safe_load(filepath:str):
    '''Returns dict or None from filepath (str)'''

    data = None
    with open(filepath) as f:
        debug(f'Opened {filepath=}')

        if 'yaml' not in globals().keys():
            error(
                'package "yaml" (via pyyaml) was not imported,' \
                + f' yet we encountered a {filepath=} file that needs to be parsed.' \
                + ' Please pip install pyyaml or similar.'
                        )

        try:
            data = yaml.safe_load(f)
        except yaml.YAMLError as e:

            # if yamllint is installed, then use it to get all errors in the .yml
            # file, instead of the single exception.
            if shutil.which('yamllint'):
                try:
                    sp_out = subprocess.run(
                        f'yamllint -d relaxed --no-warnings {filepath}'.split(),
                        capture_output=True, text=True )
                    for x in sp_out.stdout.split('\n'):
                        if x:
                            info('yamllint: ' + x)
                except:
                    pass

            if hasattr(e, 'problem_mark'):
                mark = e.problem_mark
                error(f"Error parsing {filepath=}: line {mark.line + 1}, column {mark.column +1}: {e.problem}")
            else:
                error(f"Error loading YAML {filepath=}:", e)





    return data


def yaml_safe_writer(data:dict, filepath:str) -> None:

    if filepath.endswith('.yml'):
        if 'yaml' not in globals().keys():
            warning(
                'package "yaml" (via pyyaml) was not imported, ' \
                + f' yet we would like to save {filepath=}, skipping'
            )
        else:
            with open(filepath, 'w', encoding='utf-8') as f:
                yaml.dump(data, f, default_flow_style=False, sort_keys=False)
    else:
        warning(f'{filepath=} to be written for this extension not implemented.')



def start_log(filename, force=False):
    global logfile, loglast
    if os.path.exists(filename):
        if force:
            info(f"Overwriting '{filename}', which exists, due to --force-logfile.")
        else:
            error(f"The --logfile path '{filename}' exists.  Use --force-logfile (vs --logfile) to override.")
    try:
        logfile = open( filename, 'w')
        debug(f"Opened logfile '{filename}' for writing")
    except Exception as e:
        error(f"Error opening '{filename}' for writing!")

def write_log(text, end):
    global logfile, loglast
    sw = text.startswith(f"INFO: [{progname}]")
    if (((time.time() - loglast) > 10) and
        (text.startswith(f"DEBUG: [{progname}]") or
         text.startswith(f"INFO: [{progname}]") or
         text.startswith(f"WARNING: [{progname}]") or
         text.startswith(f"ERROR: [{progname}]"))):
        dt = datetime.datetime.now().ctime()
        print(f"INFO: [{progname}] Time: {dt}", file=logfile)
        loglast = time.time()
    print(text, end=end, file=logfile)
    logfile.flush()
    os.fsync(logfile)

def stop_log():
    global logfile, loglast
    if logfile:
        debug(f"Closing logfile")
        logfile.close()
    logfile = None
    loglast = 0

atexit.register(stop_log)

def get_argparse_bool_action_kwargs():
    bool_kwargs = dict()
    x = getattr(argparse, 'BooleanOptionalAction', None)
    if x is not None:
        bool_kwargs['action'] = x
    else:
        bool_kwargs['action'] = 'store_true'
    return bool_kwargs


def process_tokens(tokens:list):
    global debug_level

    parser = argparse.ArgumentParser(prog='util.process_tokens', add_help=False, allow_abbrev=False)
    # We set allow_abbrev=False so --force-logfile won't try to attempt parsing shorter similarly
    # named args like --force, we want those to go to unparsed list.
    # For bools, support --color and --no-color with this action=argparse.BooleanOptionalAction
    # -- however Python3.8 and older does not support this, so as a workaround, use kwargs for
    #    boolean actions:
    bool_action_kwargs = get_argparse_bool_action_kwargs()

    parser.add_argument('--color', **bool_action_kwargs)
    parser.add_argument('--quiet', **bool_action_kwargs)
    parser.add_argument('--verbose', **bool_action_kwargs)
    parser.add_argument('--fancy', **bool_action_kwargs)
    parser.add_argument('--debug', **bool_action_kwargs)
    parser.add_argument('--debug-level', type=int, default=0)
    parser.add_argument('--logfile', type=str, default='')
    parser.add_argument('--force-logfile', type=str, default='')
    parser.add_argument('--no-respawn', action='store_true')
    try:
        parsed, unparsed = parser.parse_known_args(tokens + [''])
        unparsed = list(filter(None, unparsed))
    except argparse.ArgumentError:
        error(f'problem attempting to parse_known_args for {tokens=}')

    debug(f'util.process_tokens: {parsed=} {unparsed=}  from {tokens=}')

    if parsed.debug_level: set_debug_level(parsed.debug_level)
    elif parsed.debug:     set_debug_level(1)
    else:                  debug_level = 0

    if parsed.force_logfile != '':
        start_log(parsed.force_logfile, force=True)
    elif parsed.logfile != '':
        start_log(parsed.logfile, force=False)

    parsed_as_dict = vars(parsed)
    for key,value in parsed_as_dict.items():
        if value is not None:
            args[key] = value # only update with non-None values to our global 'args' dict
    return parsed_as_dict, unparsed

# ********************
# fancy support
# In fancy mode, we take the bottom fancy_lines_ lines of the screen to be written using fancy_print,
# while the lines above that show regular scrolling content (via info, debug, warning, error above).
# User should not use print() when in fancy mode

fancy_lines_ = []

def fancy_start(fancy_lines = 4, min_vanilla_lines = 4):
    global fancy_lines_
    (columns,lines) = shutil.get_terminal_size()
    if (fancy_lines < 2):
        error(f"Fancy mode requires at least 2 fancy lines")
    if (fancy_lines > (lines-min_vanilla_lines)):
        error(f"Fancy mode supports at most {(lines-min_vanilla_lines)} fancy lines, given {min_vanilla_lines} non-fancy lines")
    if len(fancy_lines_): error(f"We are already in fancy line mode??")
    for _ in range(fancy_lines-1):
        print("") # create the requisite number of blank lines
        fancy_lines_.append("")
    print("", end="") # the last line has no "\n" because we don't want ANOTHER blank line below
    fancy_lines_.append("")
    # the cursor remains at the leftmost character of the bottom line of the screen

def fancy_stop():
    global fancy_lines_
    if len(fancy_lines_): # don't do anything if we aren't in fancy mode
        # user is expected to have painted something into the fancy lines, we can't "pull down" the regular
        # lines above, and we don't want fancy_lines_ blank or garbage lines either, that's not pretty
        fancy_lines_ = []
        # since cursor is always left at the leftmost character of the bottom line of the screen, which was
        # one of the fancy lines which now has the above-mentioned "something", we want to move one lower
        print("")

def fancy_print(text, line):
    global fancy_lines_
    # strip any newline, we don't want to print that
    if text.endswith("\n"): text.rstrip()
    lines_above = len(fancy_lines_) - line - 1
    if lines_above:
        print(f"\033[{lines_above}A"+ # move cursor up
              text+f"\033[1G"+ # desired text, then move cursor to the first character of the line
              f"\033[{lines_above}B", # move the cursor down
              end="", flush=True)
    else:
        print(text+f"\033[1G", # desired text, then move cursor to the first character of the line
              end="", flush=True)
    fancy_lines_[line] = text

def print_pre():
    # stuff we do before printing any line
    if len(fancy_lines_):
        # Also, note that in fancy mode we don't allow the "above lines" to be partially written, they
        # are assumed to be full lines ending in "\n"
        # As always, we expect the cursor was left in the leftmost character of bottom line of screen
        print(f"\033[{len(fancy_lines_)-1}A"+ # move the cursor up to where the first fancy line is drawn
              f"\033[0K", # clear the old fancy line 0
              end="",flush=True)

def print_post(text, end):
    # stuff we do after printing any line
    if len(fancy_lines_):
        #time.sleep(1)
        # we just printed a line, including a new line, on top of where fancy line 0 used to be, so cursor
        # is now at the start of fancy line 1.
        # move cursor down to the beginning of the final fancy line (i.e. standard fancy cursor resting place)
        for x in range(len(fancy_lines_)):
            print("\033[0K",end="") # erase the line to the right
            print(fancy_lines_[x],flush=True,end=('' if x==(len(fancy_lines_)-1) else '\n'))
            #time.sleep(1)
        print("\033[1G", end="", flush=True)
    if logfile: write_log(text, end=end)

string_red = f"\x1B[31m"
string_green = f"\x1B[32m"
string_orange = f"\x1B[33m"
string_yellow = f"\x1B[39m"
string_normal = f"\x1B[0m"

def print_red(text, end='\n'):
    print_pre()
    print(f"{string_red}{text}{string_normal}" if args['color'] else f"{text}", end=end, flush=True)
    print_post(text, end)

def print_green(text, end='\n'):
    print_pre()
    print(f"{string_green}{text}{string_normal}" if args['color'] else f"{text}", end=end, flush=True)
    print_post(text, end)

def print_orange(text, end='\n'):
    print_pre()
    print(f"{string_orange}{text}{string_normal}" if args['color'] else f"{text}", end=end, flush=True)
    print_post(text, end)

def print_yellow(text, end='\n'):
    print_pre()
    print(f"{string_yellow}{text}{string_normal}" if args['color'] else f"{text}", end=end, flush=True)
    print_post(text, end)

def set_debug_level(level):
    global debug_level
    debug_level = level
    args['debug'] = (level > 0)
    args['verbose'] = (level > 1)
    info(f"Set debug level to {debug_level}")

# the <<d>> stuff is because we change progname after this is read in.  if we instead infer progname or
# get it passed somehow, we can avoid this ugliness / performance impact (lots of calls to debug happen)
def debug(text, level=1, start='<<d>>', end='\n'):
    if start=='<<d>>': start = f"DEBUG: " + (f"[{progname}] " if progname_in_message else "")
    if args['debug'] and (((level==1) and args['verbose']) or (debug_level >= level)):
        print_yellow(f"{start}{text}", end=end)

def info(text, start='<<d>>', end='\n'):
    if start=='<<d>>': start = f"INFO: " + (f"[{progname}] " if progname_in_message else "")
    if not args['quiet']:
        print_green(f"{start}{text}", end=end)

def warning(text, start='<<d>>', end='\n'):
    if start=='<<d>>': start = f"WARNING: " + (f"[{progname}] " if progname_in_message else "")
    args['warnings'] += 1
    print_orange(f"{start}{text}", end=end)

def error(text, error_code=-1, do_exit=True, start='<<d>>', end='\n'):
    if start=='<<d>>': start = f"ERROR: " + (f"[{progname}] " if progname_in_message else "")
    args['errors'] += 1
    print_red(f"{start}{text}", end=end)
    if do_exit:
        if args['debug']: print(traceback.print_stack())
        return exit(error_code)
    else:
        if error_code is None:
            return 0
        else:
            return abs(int(error_code))

def exit(error_code=0, quiet=False):
    if global_exit_allowed:
        if not quiet: info(f"Exiting with {args['warnings']} warnings, {args['errors']} errors")
        sys.exit(error_code)

    if error_code is None:
        return 0
    else:
        return abs(int(error_code))

def getcwd():
    try:
        cc = os.getcwd()
    except Exception as e:
        error("Unable to getcwd(), did it get deleted from under us?")
    return cc

_oc_root=None
_oc_root_set=False
def get_oc_root(error_on_fail:bool=False):
    global _oc_root
    global _oc_root_set
    '''Returns a str or False for the root directory of *this* repo.

    If environment variable OC_ROOT is set, that is used instead, otherwise attempts to use
    `git rev-parse`
    '''
    # if we've already run through here once, just return the memorized result
    if _oc_root_set: return _oc_root

    # try looking for an env var
    s = os.environ.get('OC_ROOT')
    if s:
        # TODO: is this really highest priority?
        debug(f'get_oc_root() -- returning from env: {s=}')
        _oc_root = s.strip()
    else:
        # try asking GIT
        cp = subprocess.run('git rev-parse --show-toplevel', stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                            shell=True, universal_newlines=True)
        if cp.returncode != 0:
            # TODO: at some point, address the fact that not all repos are oc_root.  Is this function asking for
            # the repo we are in?  or a pointer to the oc_root which maybe elsewhere on the system?
            if error_on_fail: error(f'Unable to get a OC_ROOT directory using git rev-parse')
            else:             info(f'Unable to get a OC_ROOT directory using git rev-parse')
        else:
            _oc_root = cp.stdout.strip()

    # there is no sense running through this code more than once
    _oc_root_set = True
    return _oc_root

def string_or_space(text, whitespace=False):
    if whitespace:
        return " " * len(text)
    else:
        return text

def sprint_time(s):
    s = int(s)
    txt = ""
    do_all = False
    # days
    if (s >= (24*60*60)): # greater than 24h, we show days
        d = int(s/(24*60*60))
        txt += f"{d}d:"
        s -= (d*24*60*60)
        do_all = True
    # hours
    if do_all or (s >= (60*60)):
        d = int(s/(60*60))
        txt += f"{d:2}:"
        s -= (d*60*60)
        do_all = True
    # minutes
    d = int(s/(60))
    txt += f"{d:02}:"
    s -= (d*60)
    # seconds
    txt += f"{s:02}"
    return txt

def safe_cp(source:str, destination:str, create_dirs:bool=False):
    try:
        # Infer if destination is a directory
        if destination.endswith('/') or os.path.isdir(destination):
            if not os.path.exists(destination) and create_dirs:
                os.makedirs(destination, exist_ok=True)
            destination = os.path.join(destination, os.path.basename(source))
        else:
            # Ensure parent directory exists if needed
            parent_dir = os.path.dirname(destination)
            if create_dirs and parent_dir and not os.path.exists(parent_dir):
                os.makedirs(parent_dir, exist_ok=True)
        # actually copy the file
        shutil.copy2(source, destination)
    except Exception as e:
        print(f"Error copying file from '{source}' to '{destination}': {e}")
    info(f"Copied {source} to {destination}")

def safe_rmdir(path):
    """Safely and reliably remove a non-empty directory."""
    try:
        # Ensure the path exists
        if os.path.exists(path):
            shutil.rmtree(path)
            info(f"Directory '{path}' has been removed successfully.")
        else:
            debug(f"Directory '{path}' does not exist.")
    except Exception as e:
        error(f"An error occurred while removing the directory '{path}': {e}")

def safe_mkdir(path : str):
    try:
        os.mkdir(path)
    except:
        try:
            os.system(f'mkdir -p {path}')
        except FileExistsError:
            pass
        else:
            error(f'unable to mkdir {path=}')

def safe_mkdirs(base : str, new_dirs : list):
    for p in new_dirs:
        safe_mkdir( os.path.join(base, p) )

def write_shell_command_file(dirpath : str, filename : str, command_lists : list, line_breaks : bool = False):
    ''' Writes new file at {dirpath}/{filename} as a bash shell command, using command_lists (list of lists)

    -- dirpath (str)        -- directory where file is written (usually eda.work/{target}_sim
    -- filename (str)       -- filename, for example compile_only.sh
    -- command_lists (list) -- list of lists. each item in the list is a list of commands (aka, how
                               subprocess.run(args) uses a list of commands.
    -- line_breaks (bool)   -- Set to True to have 1 word per line in the file followed by a line break.
                               Default False has an entry in command_lists all on a single line.

    Returns None, writes the file and chmod's it to 0x755.

    '''
    # command_lists should be a list-of-lists.
    bash_path = shutil.which('bash')
    assert type(command_lists) is list, f'{command_lists=}'
    fullpath = os.path.join(dirpath, filename)
    with open(fullpath, 'w') as f:
        f.write('#!' + bash_path + '\n\n')
        for c in command_lists:
            assert type(c) is list, f'{c=} (c must be a list) {command_lists=}'
            if len(c) > 0:
                if line_breaks:
                    # line_breaks=True - have 1 word per line, followed by \ and \n
                    sep = " \\" + "\n"
                    f.write(sep.join(c))
                    f.write(" \n")
                else:
                    # line_break=False (default) - all words on 1 line.
                    f.write(' '.join(c))
                    f.write(" \n")
            else:
                f.write("\n")
        f.write("\n")
        f.close()
        os.chmod(fullpath, 0o755)


def write_eda_config_and_args(dirpath : str, filename='eda_config.yml', command_obj_ref=None):
    if command_obj_ref is None:
        return
    fullpath = os.path.join(dirpath, filename)
    data = dict()
    for x in ['command_name', 'target', 'original_args', 'config', 'args', 'modified_args', 'files_v', 'files_sv', 'files_vhd']:
        data[x] = getattr(command_obj_ref, x, None)
    data['command_name'] = command_obj_ref.command_name

    yaml_safe_writer(data=data, filepath=fullpath)
